@@ -299,52 +299,6 @@ static gpointer tasklist_parent_class;
  */
 static GSList *tasklist_instances;
 
-/**
- * make_gtk_label_bold.
- *
- * Switches the font of label to a bold equivalent.
- * @label: The label.
- **/
-static void
-make_gtk_label_bold (GtkLabel *label)
-{
-  PangoFontDescription *font_desc;
-
-  font_desc = pango_font_description_new ();
-
-  pango_font_description_set_weight (font_desc,
-                                     PANGO_WEIGHT_BOLD);
-
-  /* This will only affect the weight of the font, the rest is
-   * from the current state of the widget, which comes from the
-   * theme or user prefs, since the font desc only has the
-   * weight flag turned on.
-   */
-  gtk_widget_modify_font (GTK_WIDGET (label), font_desc);
-
-  pango_font_description_free (font_desc);
-}
-
-static void
-make_gtk_label_normal (GtkLabel *label)
-{
-  PangoFontDescription *font_desc;
-
-  font_desc = pango_font_description_new ();
-
-  pango_font_description_set_weight (font_desc,
-                                     PANGO_WEIGHT_NORMAL);
-
-  /* This will only affect the weight of the font, the rest is
-   * from the current state of the widget, which comes from the
-   * theme or user prefs, since the font desc only has the
-   * weight flag turned on.
-   */
-  gtk_widget_modify_font (GTK_WIDGET (label), font_desc);
-
-  pango_font_description_free (font_desc);
-}
-
 static GType
 wnck_task_get_type (void) G_GNUC_CONST;
 
@@ -1865,7 +1819,8 @@ tasklist_include_window_ignoring_skip_taskbar (WnckTasklist *tasklist,
   if (active_workspace == NULL)
     return TRUE;
 
-  if (active_workspace != wnck_window_get_workspace (win))
+  if (active_workspace != wnck_window_get_workspace (win) &&
+      !wnck_window_or_transient_needs_attention (win))
     return FALSE;
 
   if (!wnck_workspace_is_virtual (active_workspace))
@@ -2095,7 +2050,7 @@ wnck_tasklist_active_window_changed (WnckScreen   *screen,
   /* FIXME: check for group modal window */
   for (active_window = wnck_screen_get_active_window (screen);
        active_window && !active_task;
-       active_window = wnck_window_get_transient(active_window))
+       active_window = wnck_window_get_transient (active_window))
     active_task = g_hash_table_lookup (tasklist->priv->win_hash,
                                        active_window);
 
@@ -2465,7 +2420,7 @@ wnck_task_popup_menu (WnckTask *task,
       text = wnck_task_get_text (win_task, FALSE);
       menu_item = gtk_image_menu_item_new_with_label (text);
       if (wnck_task_get_needs_attention (win_task)) 
-        make_gtk_label_bold (GTK_LABEL (GTK_BIN (menu_item)->child));
+        _make_gtk_label_bold (GTK_LABEL (GTK_BIN (menu_item)->child));
       g_free (text);
       
       pixbuf = wnck_task_get_icon (win_task);
@@ -2828,12 +2783,12 @@ wnck_task_update_visible_state (WnckTask *task)
       gtk_label_set_text (GTK_LABEL (task->label), text);
       if (wnck_task_get_needs_attention (task))
         {
-          make_gtk_label_bold ((GTK_LABEL (task->label)));
+          _make_gtk_label_bold ((GTK_LABEL (task->label)));
           wnck_task_queue_glow (task);
         }
       else
         {
-          make_gtk_label_normal ((GTK_LABEL (task->label)));
+          _make_gtk_label_normal ((GTK_LABEL (task->label)));
           wnck_task_stop_glow (task);
         }
       g_free (text);
@@ -2867,10 +2822,11 @@ wnck_task_state_changed (WnckWindow     *window,
       (changed_mask & WNCK_WINDOW_STATE_DEMANDS_ATTENTION) ||
       (changed_mask & WNCK_WINDOW_STATE_URGENT))
     {
-      WnckTask *win_task;
+      WnckTask *win_task = NULL;
 
-      win_task = g_hash_table_lookup (tasklist->priv->win_hash,
-						window);
+      /* FIXME: Handle group modal dialogs */
+      for (; window && !win_task; window = wnck_window_get_transient (window))
+        win_task = g_hash_table_lookup (tasklist->priv->win_hash, window);
 
       if (win_task)
 	{
@@ -2878,7 +2834,9 @@ wnck_task_state_changed (WnckWindow     *window,
 
 	  wnck_task_update_visible_state (win_task);
 	  
-	  class_group_task = g_hash_table_lookup (tasklist->priv->class_group_hash, win_task->class_group);
+	  class_group_task =
+            g_hash_table_lookup (tasklist->priv->class_group_hash,
+                                 win_task->class_group);
 
 	  if (class_group_task)
 	    wnck_task_update_visible_state (class_group_task);
@@ -3093,7 +3051,7 @@ wnck_task_create_widgets (WnckTask *task, GtkReliefStyle relief)
 
   if (wnck_task_get_needs_attention (task))
     {
-      make_gtk_label_bold ((GTK_LABEL (task->label)));
+      _make_gtk_label_bold ((GTK_LABEL (task->label)));
       wnck_task_queue_glow (task);
     }
 
